package be.iminds.ilabt.jfed.experimenter_gui.editor.views;

import be.iminds.ilabt.jfed.experimenter_gui.ExperimenterApplication;
import be.iminds.ilabt.jfed.experimenter_gui.config.JFedGuiConfig;
import be.iminds.ilabt.jfed.experimenter_gui.editor.ToolBoxFilter;
import be.iminds.ilabt.jfed.experimenter_gui.editor.ToolBoxFilterBuilder;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.*;
import be.iminds.ilabt.jfed.lowlevel.testbed_info.TestbedInfoSource;
import be.iminds.ilabt.util.jsonld.impl.PrimaryIdObject;
import javafx.collections.FXCollections;
import javafx.collections.ListChangeListener;
import javafx.event.ActionEvent;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.scene.control.Label;
import javafx.scene.layout.BorderPane;
import javafx.stage.Stage;
import org.controlsfx.control.CheckListView;
import org.jetbrains.annotations.NotNull;

import javax.annotation.Nonnull;
import java.io.IOException;
import java.net.URL;
import java.util.*;
import java.util.stream.Collectors;

/**
 * Edit dialog for a "collection of testbeds" filter
 * @param <T> Federation, Collection or TestbedCategory
 */
public class EditCollToolBoxFilterDialog<T extends PrimaryIdObject> extends BorderPane {
    private final Class<T> collClazz;

    private final JFedGuiConfig config;
    private final TestbedInfoSource testbedInfoSource;

    private final ToolBoxFilter origToolBoxFilter;
    private final ToolBoxFilterBuilder toolBoxFilter;

    @FXML private Label titleLabel;
    @FXML private Label filterHeaderLabel;
    @FXML private Label selectedCollCountLabel;
    @FXML private CheckListView<CollWrapper> collIdList;

    private Map<T, Set<Testbed>> testbedsPerColMap;
    private Map<String, Set<Testbed>> testbedsPerColIdMap;

    private boolean cancel = true;

    public EditCollToolBoxFilterDialog(Class<T> collClazz,
                                       JFedGuiConfig config,
                                       TestbedInfoSource testbedInfoSource,
                                       ToolBoxFilter toolBoxFilter) {
        this.collClazz = collClazz;

        if (!collClazz.equals(Federation.class) &&
                !collClazz.equals(TestbedCategory.class) &&
                !collClazz.equals(Organisation.class)) {
            throw new RuntimeException("Unsupported coll class " + collClazz.getSimpleName());
        }

        this.config = config;
        this.testbedInfoSource = testbedInfoSource;
        this.origToolBoxFilter = toolBoxFilter;
        this.toolBoxFilter = new ToolBoxFilterBuilder(toolBoxFilter);

        ExperimenterApplication.loadFontAwsome();
        URL location = getClass().getResource("EditCollToolBoxFilterDialog.fxml");
        assert location != null;
        FXMLLoader fxmlLoader = new FXMLLoader(location, null);
        fxmlLoader.setRoot(this);
        fxmlLoader.setController(this);

        try {
            fxmlLoader.load();
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    private Collection<T> getAllColls() {
        if (collClazz.equals(Federation.class)) {
            Set<Federation> res = config.getAllServersInVisibleResources().stream()
                    .map(Server::getTestbed)
                    .filter(Objects::nonNull)
                    .flatMap(t -> t.getFederations().stream())
                    .filter(Objects::nonNull)
                    .map(testbedInfoSource::getFederation)
                    .collect(Collectors.toSet());
            return (Collection<T>) res;
        }
        if (collClazz.equals(TestbedCategory.class)) {
            Set<TestbedCategory> res = config.getAllServersInVisibleResources().stream()
                    .map(Server::getTestbed)
                    .filter(Objects::nonNull)
                    .flatMap(t -> t.getCategories().stream())
                    .filter(Objects::nonNull)
                    .map(testbedInfoSource::getTestbedCategory)
                    .collect(Collectors.toSet());
            return (Collection<T>) res;
        }
        if (collClazz.equals(Organisation.class)) {
            Set<Organisation> res = config.getAllServersInVisibleResources().stream()
                    .map(Server::getTestbed)
                    .filter(Objects::nonNull)
                    .map(Testbed::getOrganisation)
                    .filter(Objects::nonNull)
                    .map(testbedInfoSource::getOrganisation)
                    .collect(Collectors.toSet());
            return (Collection<T>) res;
        }
        throw new RuntimeException("Unsupported coll class "+collClazz.getSimpleName());
    }

    private Map<T, Set<Testbed>> getTestbedsPerColMap() {
        if (collClazz.equals(Federation.class)) {
            Map<Federation, Set<Testbed>> res = new HashMap<>();
            for (Testbed tb: config.getAllServersInVisibleResources().stream()
                    .map(Server::getTestbed)
                    .filter(Objects::nonNull).collect(Collectors.toSet())) {
                for (Federation f: tb.getFederations()) {
                    res.computeIfAbsent(testbedInfoSource.getFederation(f), ff -> new HashSet<>()).add(tb);
                }
            }
            return (Map<T, Set<Testbed>>) res;
        }
        if (collClazz.equals(TestbedCategory.class)) {
            Map<TestbedCategory, Set<Testbed>> res = new HashMap<>();
            for (Testbed tb: config.getAllServersInVisibleResources().stream()
                    .map(Server::getTestbed)
                    .filter(Objects::nonNull).collect(Collectors.toSet())) {
                for (TestbedCategory c: tb.getCategories()) {
                    res.computeIfAbsent(testbedInfoSource.getTestbedCategory(c), cc -> new HashSet<>()).add(tb);
                }
            }
            return (Map<T, Set<Testbed>>) res;
        }
        if (collClazz.equals(Organisation.class)) {
            Map<Organisation, Set<Testbed>> res = new HashMap<>();
            for (Testbed tb: config.getAllServersInVisibleResources().stream()
                    .map(Server::getTestbed)
                    .filter(Objects::nonNull).collect(Collectors.toSet())) {
                if (tb.getOrganisation() != null)
                    res.computeIfAbsent(testbedInfoSource.getOrganisation(tb.getOrganisation()), k -> new HashSet<>()).add(tb);
            }
            return (Map<T, Set<Testbed>>) res;
        }
        throw new RuntimeException("Unsupported coll class "+collClazz.getSimpleName());
    }

    private Collection<String> getFilteredIds(@Nonnull ToolBoxFilter toolBoxFilter) {
        if (collClazz.equals(Federation.class)) {
            return toolBoxFilter.getFederationIds();
        }
        if (collClazz.equals(TestbedCategory.class)) {
            return toolBoxFilter.getCategoryNames();
        }
        if (collClazz.equals(Organisation.class)) {
            return toolBoxFilter.getOrganisationIds();
        }
        throw new RuntimeException("Unsupported coll class "+collClazz.getSimpleName());
    }

    @FXML
    private void initialize() {
        assert titleLabel != null;
        assert filterHeaderLabel != null;
        assert selectedCollCountLabel != null;
        assert collIdList != null;

        titleLabel.setText(""+collClazz.getSimpleName()+" Filters");
        filterHeaderLabel.setText(collClazz.getSimpleName()+" Filter:");

        Set<CollWrapper> allColls = getAllColls()
                .stream()
                .map(CollWrapper::new)
                .collect(Collectors.toCollection(TreeSet::new));
        Set<CollWrapper> filteredColls = allColls
                .stream()
                .filter((cw) -> getFilteredIds(origToolBoxFilter).contains(cw.getCollId()))
                .collect(Collectors.toCollection(TreeSet::new));

        testbedsPerColMap = getTestbedsPerColMap();
        testbedsPerColIdMap = testbedsPerColMap.entrySet().stream()
                .collect(Collectors.toMap(e -> getCollId(e.getKey()), Map.Entry::getValue));

        Comparator<CollWrapper> collWrapperComparable = (a, b) -> {
            boolean aFirst = filteredColls.contains(a);
            boolean bFirst = filteredColls.contains(b);
            if (aFirst != bFirst) {
                return aFirst ? -1 : 1;
            } else {
                return a.compareTo(b);
            }
        };

        List<CollWrapper> allTestbedsSorted = allColls.stream()
                .sorted(collWrapperComparable)
                .collect(Collectors.toList());

        collIdList.setItems(
                FXCollections.observableArrayList(allTestbedsSorted)
        );

        for (CollWrapper tw: filteredColls) {
            collIdList.getCheckModel().check(collIdList.getCheckModel().getItemIndex(tw));
        }

        collIdList.getCheckModel().getCheckedItems().addListener(
                (ListChangeListener<? super CollWrapper>) (e) -> updateCountLabel()
        );
        updateCountLabel();
    }

    private void updateCountLabel() {
        int count = collIdList.getCheckModel().getCheckedItems().size();
        selectedCollCountLabel.setText("("+count+" "+collClazz.getSimpleName()+" selected)");
    }

    public void onSaveAction(ActionEvent e) {
        cancel = false;
        Stage stage = (Stage) getScene().getWindow();
        stage.hide();
    }

    public void onCancelAction(ActionEvent e) {
        cancel = true;
        Stage stage = (Stage) getScene().getWindow();
        stage.hide();
    }

    public ToolBoxFilter getToolBoxFilter() {
        if (cancel)
            return origToolBoxFilter;
        else {
            Set<String> selectedCollIds = collIdList.getCheckModel().getCheckedItems()
                    .stream()
                    .map(CollWrapper::getCollId)
                    .collect(Collectors.toSet());

            if (collClazz.equals(Federation.class)) {
                toolBoxFilter.setFederationIds(selectedCollIds);
            }
            if (collClazz.equals(TestbedCategory.class)) {
                toolBoxFilter.setCategoryNames(selectedCollIds);
            }
            if (collClazz.equals(Organisation.class)) {
                toolBoxFilter.setOrganisationIds(selectedCollIds);
            }

            return toolBoxFilter.create();
        }
    }


    @Nonnull
    public String getCollId(@Nonnull T coll) {
        if (collClazz.equals(Federation.class)) {
            return ((Federation) coll).getId();
        }
        if (collClazz.equals(TestbedCategory.class)) {
            return ((TestbedCategory) coll).getName();
        }
        if (collClazz.equals(Organisation.class)) {
            return ((Organisation) coll).getId();
        }
        throw new RuntimeException("Unsupported coll class "+collClazz.getSimpleName());
    }

    private class CollWrapper implements Comparable {
        @Nonnull private final T coll;

        public CollWrapper(@Nonnull T coll) {
            this.coll = coll;
        }

        @Nonnull
        public T getColl() {
            return coll;
        }

        @Nonnull
        public String getCollId() {
            return EditCollToolBoxFilterDialog.this.getCollId(coll);
        }

        @Override
        public String toString() {
            assert coll != null;

            String suffix = testbedsPerColIdMap == null ? "" : " (contains " + testbedsPerColIdMap.getOrDefault(getCollId(), Collections.emptySet()).size()+" testbeds)";

            if (collClazz.equals(Federation.class)) {
                return (((Federation) coll).getLongName() == null ? ((Federation) coll).getId() : ((Federation) coll).getLongName())
                        + suffix;
            }
            if (collClazz.equals(TestbedCategory.class)) {
                return (((TestbedCategory) coll).getName() == null ? ((TestbedCategory) coll).getId()+"" : ((TestbedCategory) coll).getName())
                        + suffix;
            }
            if (collClazz.equals(Organisation.class)) {
                return (((Organisation) coll).getName() == null ? ((Organisation) coll).getId() : ((Organisation) coll).getName())
                        + suffix;
            }
            throw new RuntimeException("Unsupported coll class "+collClazz.getSimpleName());
        }

        @Override
        public int compareTo(@NotNull Object o) {
            assert o != null;
            return toString().toLowerCase().compareTo(o.toString().toLowerCase());
        }

        @Override
        public boolean equals(Object o) {
            return compareTo(o) == 0;
        }
    }
}
